searchData={"items":[{"type":"module","title":"FLAME","doc":"FLAME remotely executes your application code on ephemeral nodes.\n\nFLAME allows you to scale your application operations on a granular\nlevel **without rewriting your code**. For example, imagine the following function\nin your application that transcodes a video, saves the result to video storage,\nand updates the database:\n\n    def resize_video_quality(%Video{} = vid) do\n      path = \"#{vid.id}_720p.mp4\"\n      System.cmd(\"ffmpeg\", ~w(-i #{vid.url} -s 720x480 -c:a copy #{path}))\n      VideoStore.put_file!(\"videos/#{path}\", path)\n      {1, _} = Repo.update_all(from v in Video, where v.id == ^vid.id, set: [file_720p: path])\n      {:ok, path}\n    end\n\nThis works great locally and in production under no load, but video transcoding\nis necessarily an expensive CPU bound operation. In production, only a\nfew concurrent users can saturate your CPU and cause your entire application,\nweb requests, etc, to come to crawl. This is where folks typically reach for\nFaaS or external service solutions, but FLAME gives you a better way.\n\nSimply wrap your your existing code in a FLAME function and it will be executed\non a newly spawned, ephemeral node. Using Elixir and Erlang's built in distribution\nfeatures, entire function closures, including any state they close over, can be sent\nand executed on a remote node:\n\n    def resize_video_quality(%Video{} = video) do\n      FLAME.call(MyApp.FFMpegRunner, fn ->\n        path = \"#{vid.id}_720p.mp4\"\n        System.cmd(\"ffmpeg\", ~w(-i #{vid.url} -s 720x480 -c:a copy #{path}))\n        VideoStore.put_file!(\"videos/#{path}\", path)\n        {1, _} = Repo.update_all(from v in Video, where v.id == ^vid.id, set: [file_720p: path])\n        {:ok, path}\n      end)\n    end\n\nThat's it! The `%Video{}` struct in this example is captured inside the function\nand everything executes on the remotely spawned node, returning the result back to the\nparent node when it completes. Repo calls Just Work because the new node booted\nyour entire application, including the database Repo. As soon as the function is done\nexecuting, the ephemeral node is terminated. This means you can elastically scale\nyour app as load increases, and only pay for the resources you need at the time.\n\nTo support your FLAME calls, you'll need to add a named `FLAME.Pool` to your\napplication's supervision tree, which we'll discuss next.","ref":"FLAME.html"},{"type":"module","title":"Pools - FLAME","doc":"A `FLAME.Pool` provides elastic runner scaling, allowing a minimum and\nmaximum number of runners to be configured, and idle'd down as load decreases.\n\nPools give you elastic scale that maximizes the newly spawned hardware.\nAt the same time, you also want to avoid spawning unbound resources. You also\nwant to keep spawned nodes alive for a period of time to avoid the overhead\nof booting new ones before idling them down. The following pool configuration\ntakes care of all of this for you:\n\n    children = [\n      ...,\n      {FLAME.Pool,\n       name: App.FFMpegRunner,\n       min: 0,\n       max: 10,\n       max_concurrency: 5,\n       idle_shutdown_after: 30_000},\n    ]\n\nHere we add a `FLAME.Pool` to our application supervision tree, configuring\na minimum of 0 and maximum of 10 runners. This acheives \"scale to zero\" behavior\nwhile also allowing the pool to scale up to 10 runners when load increases.\nEach runner in the case will be able to execute up to 5 concurrent functions.\nThe runners will shutdown after 30 seconds of inactivity.\n\nCalling a pool is as simple as passing its name to the FLAME functions:\n\n    FLAME.call(App.FFMpegRunner, fn -> :operation1 end)\n\nYou'll also often want to enable or disable other application services based on whether\nyour application is being started as child FLAME runner or being run directly.\nSee the next `Deployment Considerations` section below for details.","ref":"FLAME.html#module-pools"},{"type":"module","title":"Deployment Considerations - FLAME","doc":"FLAME nodes effectively clone and start your entire application. This is great\nbecause all application services and dependencies are ready to go and be used to\nsupport your FLAME calls; however, You'll also often want to enable or disable\nservices based on whether your node is running as a FLAME child or not.\nFor example, there's usually no need to serve your Phoenix endpoint within a FLAME.\nYou also likely only need a single or small number of database connections instead of\nyour existing pool size.\n\nTo accomplish these you can use `FLAME.Parent.get/0` to conditionally enable or\ndisable processes in you `application.ex` file:\n\n    def start(_type, _args) do\n      flame_parent = FLAME.Parent.get()\n\n      children = [\n        ...,\n        {FLAME.Pool,\n         name: Thumbs.FFMpegRunner,\n         min: 0,\n         max: 10,\n         max_concurrency: 5,\n         idle_shutdown_after: 30_000},\n      !flame_parent && ThumbsWeb.Endpoint\n      ]\n      |> Enum.filter(& &1)\n\n      opts = [strategy: :one_for_one, name: Thumbs.Supervisor]\n      Supervisor.start_link(children, opts)\n    end\n\nHere we filter the Phoenix endpoint from being started when running as a FLAME\nchild because we have no need to handle web requests in this case.\n\nOr you can use `FLAME.Parent.get/0` to configure your database pool size:\n\n    pool_size =\n      if FLAME.Parent.get() do\n        1\n      else\n        String.to_integer(System.get_env(\"POOL_SIZE\") || \"10\")\n      end\n\n    config :thumbs, Thumbs.Repo,\n      ...,\n      pool_size: pool_size","ref":"FLAME.html#module-deployment-considerations"},{"type":"module","title":"Backends - FLAME","doc":"The `FLAME.Backend` behavior defines an interface for spawning remote\napplication nodes and sending functions to them. By default, the\n`FLAME.LocalBackend` is used, which is great for development and test\nenvironments, as you can have your code simply execute locally in most cases\nand worry about scaling the operation only in production.\n\nFor production, FLAME provides the `FLAME.FlyBackend`, which uses\n[Fly.io](https://fly.io). Because Fly deploys a containerized machine of\nyour application, a single Fly API call can boot a machine running your\nexact Docker deployment image, allowing closures to be executed across\ndistributed nodes.\n\nDefault backends can be configured in your `config/runtime.exs`:\n\n    if config_env() == :prod do\n      config :flame, :backend, FLAME.FlyBackend\n      config :flame, FLAME.FlyBackend, token: System.fetch_env!(\"FLY_API_TOKEN\")\n      ...\n    end","ref":"FLAME.html#module-backends"},{"type":"module","title":"Termination - FLAME","doc":"FLAME runs a termination process to allow remotely spawned functions time to\ncomplete before the node is terminated. This process is started automatically\nwith the library. The shutdown timeout by default is 30s, but can be configured\nin your application configuration, such as `config/runtime.exs`:\n\n    config :flame, :terminator, shutdown_timeout: :timer.seconds(10)","ref":"FLAME.html#module-termination"},{"type":"function","title":"FLAME.call/2","doc":"","ref":"FLAME.html#call/2"},{"type":"function","title":"FLAME.call/3","doc":"Calls a function in a remote runner for the given `FLAME.Pool`.","ref":"FLAME.html#call/3"},{"type":"function","title":"Options - FLAME.call/3","doc":"* `:timeout` - The timeout the caller is willing to wait for a response before an\n    exit with `:timeout`. Defaults to the configured timeout of the pool.\n    The executed function will also be terminated on the remote flame if\n    the timeout is reached.","ref":"FLAME.html#call/3-options"},{"type":"function","title":"Examples - FLAME.call/3","doc":"def my_expensive_thing(arg) do\n    FLAME.call(MyApp.Runner, fn ->\n      # I'm now doing expensive work inside a new node\n      # pubsub and repo access all just work\n      Phoenix.PubSub.broadcast(MyApp.PubSub, \"topic\", result)\n\n      # can return awaitable results back to caller\n      result\n    end)\n\nWhen the caller exits, the remote runner will be terminated.","ref":"FLAME.html#call/3-examples"},{"type":"function","title":"FLAME.cast/2","doc":"Casts a function to a remote runner for the given `FLAME.Pool`.","ref":"FLAME.html#cast/2"},{"type":"function","title":"FLAME.place_child/3","doc":"Places a child process on a remote runner for the given `FLAME.Pool`.\n\nAny child process can be placed on the remote node and it will occupy a space\nin the runner's `max_concurrency` allowance. This is useful for long running\nworkloads that you want to run asynchronously from the parent caller.\n\n*Note*: The placed child process is linked to the caller and will only survive\nas long as the caller does. This is to ensure that the child process is never\norphaned permanently on the remote node.\n\n*Note*: The child spec will be rewritten to use a temporary restart strategy\nto ensure that the child process is never restarted on the remote node when it\nexits. If you want restart behavior, you need to monitor on the parent node and\nreplace the child yourself.\n\nAccepts any child spec.","ref":"FLAME.html#place_child/3"},{"type":"function","title":"Examples - FLAME.place_child/3","doc":"{:ok, pid} = FLAME.place_child(MyRunner, {MyWorker, []})","ref":"FLAME.html#place_child/3-examples"},{"type":"behaviour","title":"FLAME.Backend","doc":"Defines a behavior for a FLAME backend.\n\nA FLAME backend is responsible for booting remote compute resources,\nconnecting them back to the parent node, and executing functions on them.\n\nThe default `FLAME.LocalBackend` simply runs your code locally, allowing\nyou to develop and test your application using `FLAME.call/3` without\nrunning an external backend.","ref":"FLAME.Backend.html"},{"type":"behaviour","title":"Messaging - FLAME.Backend","doc":"The `FLAME.Terminator` process runs on remote nodes automatically and is\nresponsible for connecting back to the parent node, notifying the parent, and\nhandling termination of remote processes started via `FLAME.call/3`, `FLAME.cast/3`,\nand `FLAME.place_child/3`. When the terminator starts on a newly booted remote\nnode, it sends the following message to the parent runner process:\n\n    {ref, {:remote_up, remote_terminator_pid}}\n\nWhere ref is the reference generated by the backend and encoded into the\n`FLAME.Parent.encode/1` string.\n\nWhen the remote terminator is going away gracefully, it sends the following message:\n\n    {ref, {:remote_shutdown, :idle}}\n\nBackend implementations can react to these messages to handle the remotely\nprovisioned instance booting up or shutting down.\n\nSee `FLAME.FlyBackend` for an example implementation of this behavior.","ref":"FLAME.Backend.html#module-messaging"},{"type":"function","title":"FLAME.Backend.handle_info/2","doc":"","ref":"FLAME.Backend.html#handle_info/2"},{"type":"callback","title":"FLAME.Backend.handle_info/2","doc":"","ref":"FLAME.Backend.html#c:handle_info/2"},{"type":"function","title":"FLAME.Backend.impl/0","doc":"","ref":"FLAME.Backend.html#impl/0"},{"type":"function","title":"FLAME.Backend.init/1","doc":"","ref":"FLAME.Backend.html#init/1"},{"type":"callback","title":"FLAME.Backend.init/1","doc":"","ref":"FLAME.Backend.html#c:init/1"},{"type":"function","title":"FLAME.Backend.remote_boot/1","doc":"","ref":"FLAME.Backend.html#remote_boot/1"},{"type":"callback","title":"FLAME.Backend.remote_boot/1","doc":"","ref":"FLAME.Backend.html#c:remote_boot/1"},{"type":"function","title":"FLAME.Backend.remote_spawn_monitor/2","doc":"","ref":"FLAME.Backend.html#remote_spawn_monitor/2"},{"type":"callback","title":"FLAME.Backend.remote_spawn_monitor/2","doc":"","ref":"FLAME.Backend.html#c:remote_spawn_monitor/2"},{"type":"function","title":"FLAME.Backend.system_shutdown/0","doc":"","ref":"FLAME.Backend.html#system_shutdown/0"},{"type":"callback","title":"FLAME.Backend.system_shutdown/0","doc":"","ref":"FLAME.Backend.html#c:system_shutdown/0"},{"type":"module","title":"FLAME.FlyBackend","doc":"The `FLAME.Backend` using [Fly.io](https://fly.io) machines.\n\nThe only required configuration is telling FLAME to use the\n`FLAME.FlyBackend` by default and the `:token` which is your Fly.io API\ntoken. These can be set via application configuration in your `config/runtime.exs`\nwithing a `:prod` block:\n\n    if config_env() == :prod do\n      config :flame, :backend, FLAME.FlyBackend\n      config :flame, FLAME.FlyBackend, token: System.fetch_env!(\"FLY_API_TOKEN\")\n      ...\n    end\n\nTo set your `FLY_API_TOKEN` secret, you can run the following commands locally:\n\n```bash\n$ fly secrets set FLY_API_TOKEN=\"$(fly auth token)\"\n```\n\nThe following backend options are supported, and mirror the\n(Fly.io machines create API)[https://fly.io/docs/machines/working-with-machines/]:\n\n  * `:cpu_kind` - The size of the runner CPU. Defaults to `\"performance\"`.\n\n  * `:cpus` - The number of runner CPUs. Defaults to  `System.schedulers_online()`\n    for the number of cores of the running parent app.\n\n  * `:memory_mb` - The memory of the runner. Must be a 1024 multiple. Defaults to `4096`.\n\n  * `:boot_timeout` - The boot timeout. Defaults to `30_000`.\n\n  * `:app` – The name of the otp app. Defaults to `System.get_env(\"FLY_APP_NAME\")`,\n\n  * `:image` – The URL of the docker image to pass to the machines create endpoint.\n    Defaults to `System.get_env(\"FLY_IMAGE_REF\")` which is the image of your running app.\n\n  * `:token` – The Fly API token. Defaults to `System.get_env(\"FLY_API_TOKEN\")`.\n\n  * `:host` – The host of the Fly API. Defaults to `\"https://api.machines.dev\"`.\n\n  * `:services` - The optional services to run on the machine. Defaults to `[]`.","ref":"FLAME.FlyBackend.html"},{"type":"function","title":"FLAME.FlyBackend.with_elapsed_ms/1","doc":"","ref":"FLAME.FlyBackend.html#with_elapsed_ms/1"},{"type":"module","title":"FLAME.Parent","doc":"Conveniences for looking up FLAME parent information.","ref":"FLAME.Parent.html"},{"type":"function","title":"FLAME.Parent.encode/1","doc":"Encodes a `%FLAME.Parent{}` struct into string.","ref":"FLAME.Parent.html#encode/1"},{"type":"function","title":"FLAME.Parent.get/0","doc":"Gets the `%FLAME.Parent{}` struct from the system environment.\n\nReturns `nil` if no parent is set.\n\nWhen booting a FLAME node, the `FLAME.Backend` is required to\nexport the `FLAME_PARENT` environment variable for the provisioned\ninstance. This value holds required information about the parent node\nand can be set using the `FLAME.Parent.encode/3` function.","ref":"FLAME.Parent.html#get/0"},{"type":"function","title":"FLAME.Parent.new/3","doc":"Returns a new `%FLAME.Parent{}` struct.\n\nThe `pid` is the parent node's `FLAME.Runner` process started by\nthe `FLAME.Pool`.","ref":"FLAME.Parent.html#new/3"},{"type":"module","title":"FLAME.Pool","doc":"Manages a pool of `FLAME.Runner`'s.\n\nPools support elastic growth and shrinking of the number of runners.","ref":"FLAME.Pool.html"},{"type":"module","title":"Examples - FLAME.Pool","doc":"children = [\n      ...,\n      {FLAME.Pool, name: MyRunner, min: 1, max: 10, max_concurrency: 100}\n    ]\n\nSee `start_link/1` for supported options.","ref":"FLAME.Pool.html#module-examples"},{"type":"module","title":"TODO - FLAME.Pool","doc":"[ ] interface to configure min/max at runtime","ref":"FLAME.Pool.html#module-todo"},{"type":"function","title":"FLAME.Pool.call/3","doc":"Calls a function in a remote runner for the given `FLAME.Pool`.\n\nSee `FLAME.call/3` for more information.","ref":"FLAME.Pool.html#call/3"},{"type":"function","title":"FLAME.Pool.cast/2","doc":"Casts a function to a remote runner for the given `FLAME.Pool`.\n\nSee `FLAME.cast/2` for more information.","ref":"FLAME.Pool.html#cast/2"},{"type":"function","title":"FLAME.Pool.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"FLAME.Pool.html#child_spec/1"},{"type":"function","title":"FLAME.Pool.place_child/3","doc":"See `FLAME.place_child/3` for more information.","ref":"FLAME.Pool.html#place_child/3"},{"type":"function","title":"FLAME.Pool.start_link/1","doc":"Starts a pool of runners.","ref":"FLAME.Pool.html#start_link/1"},{"type":"function","title":"Options - FLAME.Pool.start_link/1","doc":"* `:name` - The name of the pool, for example: `MyApp.FFMPegRunner`\n\n  * `:min` - The minimum number of runners to keep in the pool at all times.\n    For \"scale to zero\" behavior you may pass `0`. When starting as a flame child,\n    the `:min` will be forced to zero to avoid recursively starting backend resources.\n\n  * `:max` - The maximum number of runners to elastically grow to in the pool.\n\n  * `:max_concurrency` - The maximum number of concurrent executions per runner before\n    booting new runners or queueing calls. Defaults to `100`.\n\n  * `:single_use` - if `true`, runners will be terminated after each call completes.\n    Defaults `false`.\n\n  * `:backend` - The backend to use. Defaults to the configured `:flame, :backend` or\n    `FLAME.LocalBackend` if not configured.\n\n  * `:log` - The log level to use for verbose logging. Defaults to `false`.\n\n  * `:timeout` - The time to allow functions to execute on a remote node. Defaults to 30 seconds.\n    This value is also used as the default `FLAME.call/3` timeout for the caller.\n  * `:boot_timeout` - The time to allow for booting and connecting to a remote node.\n    Defaults to 30 seconds.\n\n  * `:shutdown_timeout` - The time to allow for graceful shutdown on the remote node.\n    Defaults to 30 seconds.\n\n  * `:idle_shutdown_after` - The amount of time and function check to idle a remote node\n    down after a period of inactivity. Defaults to 30 seconds. A tuple may also be passed\n    to check a specific condition, for example:\n\n        {10_000, fn -> Supervisor.which_children(MySup) == []}\n\n  * `:min_idle_shutdown_after` - The same behavior of `:idle_shutdown_after`, but applied\n    to the the `:min` pool runners. Defaults to `:infinity`.\n\n  * `:on_grow_start` - The optional function to be called when the pool starts booting a new\n    runner beyond the configured `:min`. The function receives a map with the following metadata:\n\n      * `:name` - The name of the pool\n      * `:count` - The number of runners the pool is attempting to grow to\n      * `:pid` - The pid of the async process that is booting the new runner\n\n   * `:on_grow_end` - The optional 2-arty function to be called when the pool growth process completes.\n    The 2-arity function receives either `:ok` or `{:exit, reason}`, and map with the following metadata:\n\n      * `:name` - The name of the pool\n      * `:count` - The number of runners the pool is now at\n      * `:pid` - The pid of the async process that attempted to boot the new runner\n\n  * `:on_shrink` - The optional function to be called when the pool shrinks.\n    The function receives a map with the following metadata:\n\n      * `:name` - The name of the pool\n      * `:count` - The number of runners the pool is attempting to shrink to","ref":"FLAME.Pool.html#start_link/1-options"},{"type":"module","title":"FLAME.Pool.Caller","doc":"","ref":"FLAME.Pool.Caller.html"},{"type":"module","title":"FLAME.Pool.RunnerState","doc":"","ref":"FLAME.Pool.RunnerState.html"},{"type":"module","title":"FLAME.Pool.Supervisor","doc":"","ref":"FLAME.Pool.Supervisor.html"},{"type":"function","title":"FLAME.Pool.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"FLAME.Pool.Supervisor.html#child_spec/1"},{"type":"function","title":"FLAME.Pool.Supervisor.init/1","doc":"","ref":"FLAME.Pool.Supervisor.html#init/1"},{"type":"function","title":"FLAME.Pool.Supervisor.start_link/1","doc":"","ref":"FLAME.Pool.Supervisor.html#start_link/1"},{"type":"module","title":"FLAME.Pool.WaitingState","doc":"","ref":"FLAME.Pool.WaitingState.html"},{"type":"module","title":"FLAME.Terminator.Caller","doc":"","ref":"FLAME.Terminator.Caller.html"},{"type":"module","title":"FLAME.Terminator.Supervisor","doc":"","ref":"FLAME.Terminator.Supervisor.html"},{"type":"function","title":"FLAME.Terminator.Supervisor.child_placement_sup_name/1","doc":"","ref":"FLAME.Terminator.Supervisor.html#child_placement_sup_name/1"},{"type":"function","title":"FLAME.Terminator.Supervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"FLAME.Terminator.Supervisor.html#child_spec/1"},{"type":"function","title":"FLAME.Terminator.Supervisor.init/1","doc":"","ref":"FLAME.Terminator.Supervisor.html#init/1"},{"type":"function","title":"FLAME.Terminator.Supervisor.start_link/1","doc":"","ref":"FLAME.Terminator.Supervisor.html#start_link/1"}],"content_type":"text/markdown"}